unit Zett;

{Extension of the FPC Run Time Library
 Part of the z4free library package.
 
 For use with the Free Pascal Compiler (http://www.freepascal.org/) or the
 Lazarus project (http://www.lazarus.freepascal.org/)}

{Copyright (C) 2007 Robert H. Mangl

 Zett is free software: you can redistribute it and/or modify it under the terms
 of the GNU General Public License as published by the Free Software Foundation,
 either version 3 of the License, or (at your option) any later version.

 Zett is distributed in the hope that it will be useful, but WITHOUT ANY
 WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 PARTICULAR PURPOSE.  See the GNU General Public License for more details.

 You should have received a copy of the GNU General Public License along with
 Zett.  If not, see <http://www.gnu.org/licenses/>.}

 {$mode objfpc}{$H+}

 interface

  uses Classes, SysUtils, Math;

  const
  
   Ln10 = Ln( 10);
   MSBCardIndex = 8 * SizeOf( Cardinal) - 1; // index of msb
   MSBCardinal = 1 shl MSBCardIndex; // value of most significant bit
   MaxCardinal = (MSBCardinal - 1) * 2 + 1;
   MinInteger  = 1 shl (8 * SizeOf( Integer) - 1);
   MaxInteger  = MinInteger - 1;
   MinSizeInt  = 1 shl (8 * SizeOf( SizeInt) - 1);
   MaxSizeInt  = MinSizeInt - 1;

   // utf-8 status
   
    usValid           = 0;
    usNoSequence      = 1;
    usOvercoded       = 2;
    usReversedBOM     = 3;
    usIllByteSequence = 4;

   URLEncodes : TSysCharSet = [' ', '%', '/', ':', #$80 .. #$ff];

   // character identifiers
    chrColon = ':';
    chrComma = ',';
    chrEqual = '=';
    chrSemiColon = ';';
    chrTab = #$09;
    chrVertiBar = '|';

  type
  
   CPAction = procedure( FileName: string; Params: Pointer; var Stop: Boolean);
   CPFunction = function( X: Float): Float;
   TWrapStyle = (wsNone, wsLeft, wsCenter, wsRight, wsJustify);
   TRational = record
     IsRational: Boolean;
     I: Float;
     P, Q: Integer;
    end;
   TMsgRec = record
     ID: shortstring;
     Data: Pointer
    end;
   TBoundsRec = record // for use with forms and controls
     Left: Integer;
     Top: Integer;
     Width: Integer;
     Height: Integer
    end;
   EZett = class( Exception);
   EZUTF8 = class( EZett);
   TLangMode = (lmCountry, lmEncoding);
   TLangModeSet = set of TLangMode;
   
  var
   TabWidth : Byte = 8;
   LineWidth : Byte = 80;

  // system and sysutils
  procedure CreateDirectory( Directory: string); {makes sure you get an error
   if dir doesn't exist by switching $I on. leaves $I that way.}
  function CurrentException: Exception;
  procedure EraseFiles( Directory, Pattern: string;
                        ShallTraverse: Boolean = False);
   {erases all files within Directory that match the pattern. if Traverse is set
    to True, EraseFile will search and destroy in all subirecturies, down to the
    last}
  function GetApplicationDir: string;
  function GetConfigDir( Global: Boolean): string;
  function GetFullUserName: string;
  function GetSysLang( LangModes: TLangModeSet = []): string;
  function GetLoginName: string;
  function GetRootDir( Dir: string): string;
  function MagicNumber( IDString: string): Cardinal;
  function MagicString( IDNumber: Cardinal): string;
  procedure NoOperation;
  function RelativeLoc( BaseDir, Loca: string): string; {as
   SysUtils.ExtractRelativepath but better! expects full paths, but might work
   with some non-absolute ones}
  procedure SplitLoc( Loc: string; var Drive, Path, Name, Ext: string);
  procedure Swap( var A, B: Float);
  procedure Swap( var A, B: Integer);
  function TestPattern( Pattern, TestedBits, ClearedBits: Cardinal): Boolean;
  function TestPattern( Pattern, TestedBits: Cardinal): Boolean;
  function Traverse( RootDir, Pattern: string;
                     Action: CPAction; Params: Pointer): string;
  function XEQT( Parent: TComponent; Command: string): string;

  // math
  procedure CancelDown( var P, Q: Integer);
  function CuDiGauss( La: Float): Float;
  procedure DecMantExp( F: Float; out Mantissa: Float; out Exponent: Integer);
  function DecPlaces: SizeInt;
  function DensGauss( La: Float): Float;
  function FloatWidth: SizeInt;
  function Floor( R: Float): Float;
  function GCD( A, B: Integer): Integer;
  function Integrate( F: CPFunction; A, B, Ep: Float): Float;
  function LCM( A, B: Integer): Integer;
  function Lg( X: Float): Float;
  function MaxDecimals: SizeInt;
  function MaxPreciseFloat: Float;
  function MinEp: Float;
  function Rationalize( R: Float): TRational;
  function RoundSig( F: Float; Digits: SizeInt): Float;
  function RoundTo( F: Float; Precision: SizeInt): Float;
  function Solve( F: CPFunction; X1, X2, EpX, EpY: Float): Float;
  
  // strutils
  function CardinalVal( S: string): Cardinal;
  function DecodeUTF8( UTF8: string): Cardinal;
  function DecodeUTF8( UTF8: string; Ignore: Boolean): Cardinal;
  function DecodeUTF8( UTF8: string; out Status: Byte): Cardinal;
  function EncodeUTF8( Unicode: Cardinal): string;
  function DequeueUTF8( var S: string): string;
  function DiffPos( S1, S2: string): SizeInt; // returns first different byte
  function DirForm( Dir: string): string;
  function FloatVal( S: string): Float;
  function FromStr( S: string; P: SizeInt): string;
  function IntegerVal( S: string): Integer;
  function ProcDirSep( D: string): string; {processes a string so that it can be
   used as directory: converts separators to local system customs and appends
   one if necessary.}
  function ReplaceChar( S: string; C, R: Char): string;
  function RPosSetEx( Chars: TSysCharSet; S: string; Offset: SizeInt): SizeInt;
  function SizeIntVal( S: string): SizeInt;
  function SubStr( S: string; First, Last: SizeInt): string;
  function TrimFloat( F: Float; Width, Decimals: Cardinal): string;
  function URLDecode( S: string): string;
  function URLEncode( S: string): string;
  function URLEncode( S: string; Extra: TSysCharSet;
                      ExtraOnly: Boolean = False): string;
  function Wrap( S: string; WrapStyle : TWrapStyle = wsNone;
                 UTF8 : Boolean = True): string;

 implementation
 
  uses StrUtils, Process
       {$IFDEF WINDOWS}
        , ZWindows
       {$ENDIF}
       ;
 
  procedure CancelDown( var P, Q: Integer);
   var
    D: Integer;
   begin
    D := GCD( P, Q);
    P := P div D;
    Q := Q div D
   end;

  function CardinalVal( S: string): Cardinal;
   var
    F: Float;
   begin
    F := FloatVal( S);
    if F < 0
     then raise EZett.Create( 'CardinalVal - argument < 0');
    if F > MaxCardinal
     then raise EZett.Create( 'CardinalVal - argument > ' +
                                IntToStr( MaxCardinal));
    CardinalVal := Round( F);
   end;

  procedure CreateDirectory( Directory: string);
   begin
     {$I+}
     MkDir( Directory)
    end;

  function CurrentException: Exception;
   begin
    if RaiseList <> nil
     then CurrentException := Exception( RaiseList^.fObject)
     else CurrentException := nil
   end;

  function CuDiGauss( La: Float): Float;
   var
    I         , // invariant part of the summands
    OldV, NewV, // variant part of the summands
    SqrLa,      // square of La
    SiV,        // sum of all variant parts
    Ep: Float;  // maximum summand to stop
    D    : Cardinal; // denominator of V
   begin
     I := DensGauss( La);
     Ep := MinEp / I;
     I := I * Sign( La);
     NewV := Abs( La);
     SiV := NewV;
     SqrLa := Sqr( La);
     D := 1;
     repeat
       OldV := NewV;
       Inc( D, 2);
       NewV := OldV * SqrLa / D;
       SiV := SiV + NewV
      until (NewV <= OldV) and (NewV <= Ep);
     CuDiGauss := I * SiV + 0.5
    end;

  procedure DecMantExp( F: Float; out Mantissa: Float; out Exponent: Integer);
   var
    LgF, LgMantissa, Sgn: Float;
   begin
    if F = 0
     then begin
           Mantissa := 0;
           Exponent := 0
          end
     else begin
           Sgn := Sign( F);
           F := Abs( F);
           LgF := Lg( F);
           Exponent := Trunc( Floor( LgF));
           LgMantissa := LgF - Exponent;
           Mantissa := Sgn * 10 ** LgMantissa
          end
   end;
   
  function DecodeUTF8( UTF8: string): Cardinal;
   begin
    DecodeUTF8 := DecodeUTF8( UTF8, False)
   end;
   
  function DecodeUTF8( UTF8: string; Ignore: Boolean): Cardinal;
   var
    Status: Byte;
   procedure Err( Msg: string);
    begin
     raise EZUTF8.Create( Msg)
    end;
   begin
    DecodeUTF8 := DecodeUTF8( UTF8, Status);
    if not Ignore
     then case Status of
            usNoSequence     : Err( 'no utf-8 sequence at all!');
            usOvercoded      : Err( 'overcoded');
            usReversedBOM    : Err( 'reversed bom encountered');
            usIllByteSequence: Err( 'illegal byte sequence')
           end
   end;
   
  function DecodeUTF8( UTF8: string; out Status: Byte): Cardinal;
   var
    StartByte, // start byte of a utf-8 sequence
    TrailByte, // trailing bytes of a utf-8 sequence
    Len,       // actual langth of the passed utf-8 sequence
    ValidLen,  // valid length of an utf-8 sequence according to start byte
    I: Byte;   // index of counting loop
   begin
    DecodeUTF8 := 0;
    if UTF8 = ''
     then Status := usNoSequence
     else begin
           Status := usValid;
           StartByte := Ord( UTF8[ 1]);
           Len := Length( UTF8);
           case StartByte of
             $00 .. $7f: begin
                          if Len = 1
                           then DecodeUTF8 := StartByte
                           else Status := usIllByteSequence;
                          Exit
                         end;
             $c0 .. $df: begin
                          ValidLen := 2;
                          Result := StartByte and $1F
                         end;
             $e0 .. $ef: begin
                          ValidLen := 3;
                          Result := StartByte and $0F
                         end;
             $f0 .. $f4: begin
                          ValidLen := 4;
                          Result := StartByte and $07
                         end;
             else        begin
                          Status := usIllByteSequence;
                          Exit
                         end
            end;
           if Len <> ValidLen
            then begin
                  Status := usIllByteSequence;
                  Exit
                 end;
           for I := 2 to Len do
            begin
             TrailByte := Ord( UTF8[ I]);
             if TestPattern( TrailByte, $c0, $40)
              then Result := (Result shl 6) or (TrailByte and $3f)
              else begin
                    Status := usIllByteSequence;
                    Exit
                   end
            end;
           if Result < $80
            then begin
                  Status := usOvercoded;
                  Exit
                 end;
           if Result = $fffe
            then begin
                  Status := usReversedBOM;
                  Exit
                 end;
           if (Result > $D7ff) and (Result < $e000)
            then begin
                  Status := usIllByteSequence;
                  Exit
                 end
          end
   end;
   
  function DecPlaces: SizeInt;
   var
    S: string;
    E, P: SizeInt;
    F: Float = 0;
   begin
    Str( F, S);
    E := Pos( 'E', S);
    P := Pos( '.', S);
    DecPlaces := E - P - 1
   end;

  function DequeueUTF8( var S: string): string;
   var
    S1: Byte;
   begin
    DequeueUTF8 := '';
    if S <> ''
     then repeat
           Result := Result + S[ 1];
           S := FromStr( S, 2);
           if S = ''
            then Break;
           S1 := Ord( S[ 1])
          until TestPattern( S1, $80, $80) or TestPattern( S1, $c0)
   end;

  function DiffPos( S1, S2: string): SizeInt;
   var
    P: SizeInt;
   begin
    P := 1;
    while (P <= Length( S1)) and (P <= Length( S2)) and (S1[ P] = S2[ P]) do
     Inc( P);
    DiffPos := P
   end;

  function DirForm( Dir: string): string;
   begin
     DoDirSeparators( Dir);
     if RightStr( Dir, 1) <> DirectorySeparator
      then Dir := Dir + DirectorySeparator;
     DirForm := Dir
    end;

  function EncodeUTF8( Unicode: Cardinal): string;
   var
    UTF8: string;
    HeadByte, TrailByte: Byte;
   begin
    if Unicode > $0010FFFD
     then EncodeUTF8 := ''
     else if Unicode < $80
           then EncodeUTF8 := Chr( Unicode)
           else begin
                 UTF8 := '';
                 HeadByte := $80;
                 repeat
                  TrailByte := (Unicode and $3F) or $80;
                  Unicode := Unicode shr 6;
                  UTF8 := Chr( TrailByte) + UTF8;
                  HeadByte := (HeadByte shr 1) or $80
                 until Unicode < 8;
                 HeadByte := HeadByte or Unicode;
                 UTF8 := Chr( HeadByte) + UTF8;
                 EncodeUTF8 := UTF8
                end
   end;

  function DensGauss( La: Float): Float;
   begin
     DensGauss := Exp( -Sqr( La) / 2) / SqRt( 2 * Pi)
    end;

  function FloatWidth: SizeInt;
   var
    F: Float = -1;
    S: string;
   begin
    Str( F, S);
    FloatWidth := Length( S)
   end;

  function FloatVal( S: string): Float;
   var
    E: SizeInt;
    V: Float;
   begin
    if S = ''
     then FloatVal := 0
     else repeat
           Val( S, V, E);
           case E of
             0  : FloatVal := V;
             1  : begin
                   FloatVal := 0;
                   E := 0
                  end
             else S := LeftStr( S, E - 1)
            end
          until E = 0
    end;

  function Floor( R: Float): Float;
   begin
    Result := Int( R);
    if (Frac( R) <> 0) and (R < 0)
     then Result := Result - 1
   end;
   
  function FromStr( S: string; P: SizeInt): string;
   begin
    FromStr := Copy( S, P, Length( S) - P + 1)
   end;

  function GCD( A, B: Integer): Integer;
   var
    R: Integer;
   begin
    case Ord( A <> 0) + 2 * Ord( B <> 0) of
      0: raise EInvalidArgument.Create( 'gcd: a = b = 0');
      1: GCD := Abs( A);
      2: GCD := Abs( B);
      3: begin
          A := Abs( A);
          B := Abs( B);
          repeat
           R := A mod B;
           if R <> 0
            then begin
                  A := B;
                  B := R
                 end
          until R = 0;
          GCD := B
         end
     end
   end;

  procedure EraseFilesAction( FileName: string; Params: Pointer;
                              var Stop: Boolean);
   begin
     DeleteFile( FileName)
    end;

  procedure EraseFiles( Directory, Pattern: string; ShallTraverse: Boolean);
   const
    NonDir = faAnyFile and not faDirectory;
   var
    OldDir: string;
    SearchRec: TSearchRec;
    Found: Boolean;
   begin
     OldDir := GetCurrentDir;
     if Directory = ''
      then Directory := '.';
     if Pattern = ''
      then Pattern := '*';
     if not SetCurrentDir( Directory)
      then Exit;
     if not ShallTraverse
      then begin
        Found := FindFirst( Pattern, NonDir, SearchRec) = 0;
        while Found do begin
          DeleteFile( Searchrec.Name);
          Found := FindNext( SearchRec) = 0
         end
       end
      else Traverse( Directory, Pattern, @EraseFilesAction, nil)
    end;

  function GetApplicationDir: string;
   var
    LastDelimPos: SizeInt;
   begin
    LastDelimPos := RPos( PathDelim, ParamStr( 0));
    GetApplicationDir := LeftStr( ParamStr( 0), LastDelimPos)
   end;

  function GetConfigDir( Global: Boolean): string;
   var
    Redundant, AppConfigDir: string;
    RedundPos: SizeInt;
   begin
    if VendorName <> ''
     then Redundant := VendorName
     else Redundant := ApplicationName;
    AppConfigDir := GetAppConfigDir( Global);
    RedundPos := RPos( Redundant, AppConfigDir);
    if RedundPos = 0
     then GetConfigDir := ProcDirSep( AppConfigDir)
     else GetConfigDir := ProcDirSep( LeftStr( AppConfigDir, RedundPos - 1))
   end;

  function GetFullUserName: string;
   var
    LoginName, PasswdLine, FullUserName: string;
   begin
    LoginName := GetLoginName;
    {$IFDEF WINDOWS}
      GetFullUserName := LoginName
     {$ELSE}
      PasswdLine := XEQT( nil, 'grep ' + LoginName + ' /etc/passwd');
      FullUserName := ExtractWord( 5, PasswdLine, [':']);
      if FullUserName <> ''
       then GetFullUserName := FullUserName
       else GetFullUserName := LoginName
    {$ENDIF}
   end;

  function GetSysLang( LangModes: TLangModeSet = []): string;
   var
    P: SizeInt;
    SysLang, Lang, Country, Encoding: string;
    T: Char;
   begin
     {$IFDEF WINDOWS}
       SysLang := GetWinSysLang;
      {$ELSE}
       SysLang := GetEnvironmentVariable( 'LANG');
     {$ENDIF}
     P := Pos( '.', SysLang);
     if P <> 0
      then begin
        Encoding := FromStr( SysLang, P + 1);
        SysLang := LeftStr( SysLang, P - 1)
       end
      else Encoding := '';
     P := Pos( '_', SysLang);
     if P <> 0
      then begin
        Lang := LeftStr( SysLang, P - 1);
        Country := FromStr( SysLang, P + 1)
       end;
     GetSysLang := Lang;
     if lmCountry in LangModes
      then Result := Result + '_' + Country;
     if lmEncoding in LangModes
      then Result := Result + '.' + Encoding
    end;

  function GetLoginName: string;
   begin
    {$IFDEF WINDOWS}
      GetLoginName := GetEnvironmentVariable( 'USERNAME')
     {$ELSE}
      GetLoginName := XEQT( nil, 'whoami')
    {$ENDIF}
   end;

  function GetRootDir( Dir: string): string;
   begin
    if Dir = ''
     then Dir := ParamStr( 0)
     else if Dir = '.'
           then Dir := GetCurrentDir;
    Result := ExtractFileDrive( Dir) + '/';
    DoDirSeparators( Result);
   end;
   
  function IntegerVal( S: string): Integer;
   var
    F: Float;
   begin
    F := FloatVal( S);
    if F < 0
     then raise EZett.Create( 'IntegerVal - argument < 0');
    if F > MaxInteger
     then raise EZett.Create( 'IntegerVal - argument > ' +
                                IntToStr( MaxInteger));
    IntegerVal := Round( F);
   end;

  function Integrate( F: CPFunction; A, B, Ep: Float): Float;
   var
    H, HOld, I, IOld, S, Sgn, X: Float;
    D: Boolean;
   begin
    if A = B
     then begin
           Integrate := 0;
           Exit
          end;
    if B < A
     then begin
           Swap( A, B);
           Sgn := -1
          end
     else Sgn := 1;
    if Ep = 0
     then Ep := MinEp;
    H := (B - A) / 2;
    I := H / 3 * (F( A) + 4 * F( A + H) + F( B));
    repeat
     HOld := H;
     IOld := I;
     H := H / 3;
     S := 0;
     X := A + H;
     D := True;
     repeat
      if D
       then S := S + 2 * F( X) + F( X + H)
       else S := S + F( X) + 2 * F( X + H);
      D := not D;
      X := X + HOld
     until X > B;
     I := (IOld + 2 * H * S) / 3;
    until Abs( I - IOld) < Ep;
    Integrate := Sgn * RoundTo( I, Round( -Lg( Ep)))
   end;
   
  function LCM( A, B: Integer): Integer;
   begin
    if (A = 0) or (B = 0)
     then LCM := 0
     else LCM := (A * B) div GCD( A, B)
   end;
   
  function Lg( X: Float): Float;
   begin
    Lg := Ln( X) / Ln10
   end;
   
  function MagicNumber( IDString: string): Cardinal;
   begin
    if Length( IDString) <> 4
     then raise EZett.Create( 'magic string "' + IDString +
                                '" is not 4 bytes long');
    Move( IDString[ 1], Result, 4)
   end;
   
  function MagicString( IDNumber: Cardinal): string;
   begin
    SetLength( MagicString, 4);
    Move( IDNumber, Result[ 1], 4)
   end;

  function MaxDecimals: SizeInt;
   var
    F: Float = 1;
    S: string;
    D, E: SizeInt;
   begin
    Str( F, S);
    D := Pos( '.', S);
    E := Pos( 'E', S);
    MaxDecimals := E - D
   end;
   
  function MaxPreciseFloat: Float;
   begin
    MaxPreciseFloat := (10 ** MaxDecimals - 1) * 10 + 9
   end;
   
  function MinEp: Float;
   var
    S: string;
   begin
    S := '1E-' + IntToStr( DecPlaces);
    Val( S, MinEp)
   end;
   
  procedure NoOperation; begin end;
  
  function Rationalize( R: Float): TRational;
   var
    Frc: Float;
    Cyclic, ACyclic, CDiv, ADiv: Integer;
    FracStr, CyclStr, AcycStr: string;
   procedure Extract;
    var
     PPos, EPos, E: SizeInt;
     Expo: Integer;
     S, ExpStr: string;
    begin
     Str( Frc, S);
     S := TrimLeft( S);
     PPos := Pos( '.', S);
     Delete( S, PPos, 1);
     EPos := Pos( 'E', S);
     FracStr := LeftStr( S, EPos - 2);
     ExpStr := FromStr( S, Epos + 1);
     Val( ExpStr, Expo, E);
     FracStr := DupeString( '0', -(1 + Expo)) + FracStr
    end; // Extract
   procedure Separate;
    var
     AL, CL, FL, LC, TP: SizeInt;
     C1, C2: string;
     IsRational: Boolean;
    begin
     FL := Length( FracStr);
     for AL := 0 to FL do
      for CL := 1 to FL do
       begin
        CyclStr := Copy( FracStr, AL + 1, CL);
        TP := AL + CL + 1;
        while TP < FL do
         begin
          C1 := Copy( FracStr, TP, CL);
          LC := Length( C1);
          if LC = CL
           then C2 := CyclStr
           else C2 := Copy( CyclStr, 1, LC);
          IsRational := C1 = C2;
          if not IsRational
           then Break;
          Inc( TP, CL)
         end;
        if IsRational
         then begin
               AcycStr := Copy( FracStr, 1, AL);
               Exit
              end
       end;
     CyclStr := '';
     AcycStr := ''
    end; // Separate
   procedure Numbers;
    var
     E: SizeInt;
    begin
     if CyclStr <> ''
      then begin
            Val( CyclStr, Cyclic, E);
            CDiv := Trunc( Power( 10, Length( CyclStr))) - 1;
            if Cyclic <> 0
             then CancelDown( Cyclic, CDiv);
           end
      else CDiv := 0;
     if AcycStr <> ''
       then begin
             Val( AcycStr, Acyclic, E);
             ADiv := Trunc( Power( 10, Length( AcycStr)));
             if Acyclic <> 0
              then CancelDown( Acyclic, ADiv)
            end
       else ADiv:= 0
    end;
   begin // - Rationalize
    with Result do
     begin
      I := Floor( R);
      Frc := R - I;
      P := 0;
      Q := 1;
      Extract;
      Separate;
      Numbers;
      case Ord( CDiv <> 0) * 2 + Ord( ADiv <> 0) of
        0: begin
            IsRational := False;
            I := I + Frc;
            Exit
           end;
        1: begin
            P := Acyclic;
            Q := ADiv
           end;
        2: begin
            P := Cyclic;
            Q := CDiv
           end;
        3: begin
            P := Acyclic * CDiv + Cyclic;
            Q := ADiv * CDiv
           end
       end;
      IsRational := True;
      if I < 0
       then begin
             I := I + 1;
             P := P - Q
            end;
      CancelDown( P, Q)
     end
   end;
   
  function RelativeLoc( BaseDir, Loca: string): string;
   const
    StepUp = '..' + DirectorySeparator;
   var
    BDrive, BPath, BName, BExt: string;
    LDrive, LPath, LName, LExt: string;
    UpSteps: string;
    P: SizeInt;
   begin
    SplitLoc( BaseDir, BDrive, BPath, BName, BExt);
    if BName + BExt <> ''
     then BPath := BPath + BName + BExt + DirectorySeparator;
    SplitLoc( Loca   , LDrive, LPath, LName, LExt);
    if BDrive <> LDrive
     then Exit( Loca);
    UpSteps := '';
    P := DiffPos( BPath, LPath);
    LPath := FromStr( LPath, P);
    repeat
     BPath := FromStr( BPath, P);
     if BPath = ''
      then Break;
     UpSteps := UpSteps + StepUp;
     P := Pos( DirectorySeparator, BPath) + 1;
    until False;
    RelativeLoc := UpSteps + LPath + LName + LExt
   end;

  function ProcDirSep( D: string): string;
   begin
     DoDirSeparators( D);
     if RightStr( D, 1) <> DirectorySeparator
      then D := D + DirectorySeparator;
     ProcDirSep := D
    end;

  function ReplaceChar( S: string; C, R: Char): string;
   var
    P, L: SizeInt;
   begin
    L := Length( S);
    if L <> 0
     then for P := 1 to L do
           if S[ P] = C
            then S[ P] := R;
    ReplaceChar := S
   end;

  function RoundSig( F: Float; Digits: SizeInt): Float;
   var
    Precision: SizeInt;
    S: string;
   begin
    if F = 0
     then RoundSig := 0
     else begin
           if Digits < 1
            then raise EInvalidArgument.Create( 'roundsig: digits < 1');
           if Digits > DecPlaces
            then Digits := DecPlaces;
           Precision := Digits + 1;
           Str( F: FloatWidth: Precision, S);
           Val( S, RoundSig)
          end
   end;
   
  function RoundTo( F: Float; Precision: SizeInt): Float;
   var
    S, M: Float;
   begin
    M := 10 ** Precision;
    S := Sign( F);
    F := Abs( F);
    F := Int( F * M + 0.5) / M;
    RoundTo := S * F
   end;

  function RPosSetEx( Chars: TSysCharSet; S: string; Offset: SizeInt): SizeInt;
   begin
    if (S = '') or (Chars = []) or (Offset < 1)
     then Exit( 0);
    if Offset > Length( S)
     then Offset := Length( S);
    while not (S[ Offset] in Chars) and (Offset > 0) do
     Dec( Offset);
    RPosSetEx := Offset
   end;

  function SizeIntVal( S: string): SizeInt;
   var
    F: Float;
   begin
    F := FloatVal( S);
    if F < 0
     then raise EZett.Create( 'SizeIntVal - argument < 0');
    if F > MaxSizeInt
     then raise EZett.Create( 'SizeIntVal - argument > ' +
                                IntToStr( MaxSizeInt));
    SizeIntVal := Round( F);
   end;

  function Solve( F: CPFunction; X1, X2, EpX, EpY: Float): Float;
   var
    DeX, X, Y, Y1, Y2: Float;
   begin
    if X1 = X2
     then raise EInvalidArgument.Create( 'solve: a = b');
    if X2 < X1
     then Swap( X1, X2);
    Y1 := F( X1);
    if Abs( Y1) < EpY
     then begin
           Solve := X1;
           Exit
          end;
    Y2 := F( X2);
    if Abs( Y2) < EpY
     then begin
           Solve := X2;
           Exit
          end;
    if Sign( Y1) = Sign( Y2)
     then raise EInvalidArgument.Create( 'solve: sign a = sign b');
    if EpX <= 0
     then EpX := MinEp;
    if EpY <= 0
     then EpY := MinEp;
    repeat
     X := (X1 + X2) / 2;
     Y := F( X);
     if Sign( Y) = Sign( Y1)
      then begin
            X1 := X;
            Y1 := Y
           end
      else X2 := X;
     DeX := X2 - X1
    until (Abs( Y) < EpY) and (DeX < EpX);
    Solve := RoundTo( X, Round( -Lg( EpX)))
   end;
   
  procedure SplitLoc( Loc: string; var Drive, Path, Name, Ext: string);
   begin
    DoDirSeparators( Loc);
    Drive := ExtractFileDrive( Loc);
    Path := ExtractFilePath( Loc);
    if (Drive <> '') and (Path <> '')
     then Path := FromStr( Path, Length( Drive) + 1);
    if Path = ''
     then Path := DirectorySeparator;
    if Path[ 1] <> DirectorySeparator
     then Path := DirectorySeparator + Path;
    if RightStr( Path, 1) <> DirectorySeparator
     then Path := Path + DirectorySeparator;
    Name := ExtractFileName( Loc);
    Ext := ExtractFileExt( Loc);
    if (Name <> '') and (Ext <> '')
     then Name := LeftStr( Name, Length( Name) - Length( Ext))
   end;

  function SubStr( S: string; First, Last: SizeInt): string;
   begin
    SubStr := Copy( S, First, Last - First + 1)
   end;
   
  procedure Swap( var A, B: Float);
   var
    C: Float;
   begin
    C := A;
    A := B;
    B := C
   end;

  procedure Swap( var A, B: Integer);
   var
    C: Integer;
   begin
    C := A;
    A := B;
    B := C
   end;
   
  function TestPattern( Pattern, TestedBits, ClearedBits: Cardinal): Boolean;
   begin
    ClearedBits := ClearedBits and TestedBits;
    TestPattern := ((Pattern and TestedBits) xor ClearedBits) = TestedBits
   end;
   
  function TestPattern( Pattern, TestedBits: Cardinal): Boolean;
   begin
    TestPattern := (Pattern and TestedBits) = TestedBits
   end;

  function Traverse( RootDir, Pattern: string;
                     Action: CPAction; Params: Pointer): string;
   const
    NonDir = faAnyFile and not faDirectory;
   var
    OldDir, FullName: string;
    SearchRec: TSearchRec;
    Found, Stop: Boolean;
   begin
    OldDir := GetCurrentDir;
    if RootDir = ''
     then RootDir := GetRootDir( '.');
    if SetCurrentDir( RootDir)
     then begin
           Traverse := '';
           if Pattern = ''
            then Pattern := '*';
           Found := FindFirst( Pattern, NonDir, SearchRec) = 0;
           Stop := False;
           while Found do
            begin
             FullName := ExpandFileName( SearchRec.Name);
             Action( FullName, Params, Stop);
             if Stop
              then begin
                    Traverse := FullName;
                    Break
                   end;
             Found := FindNext( SearchRec) = 0
            end;
           FindClose( SearchRec);
           if not Stop
            then begin
                  Found := FindFirst( '*', faDirectory, SearchRec) = 0;
                  while Found do
                   begin
                    with SearchRec do
                     if (Attr and faDirectory <> 0) and
                        (Name <> '.') and (Name <> '..')
                      then begin
                            FullName :=
                             Traverse( Name, Pattern, Action, Params);
                            if FullName <> ''
                             then begin
                                   Traverse := FullName;
                                   Break
                                  end
                           end;
                    Found := FindNext( SearchRec) = 0
                   end;
                  FindClose( SearchRec)
                 end;
           SetCurrentDir( OldDir);
          end
     else Traverse :='.'
   end;

  function TrimFloat( F: Float; Width, Decimals: Cardinal): string;
   var
    T: string;
   begin
    Str( F: Width: Decimals, T);
    T := TrimLeft( T);
    T := TrimRightSet( T, ['0']);
    if RightStr( T, 1) = '.'
     then T := LeftStr( T, Length( T) - 1);
    TrimFloat := T
   end;

  function URLDecode( S: string): string;
   var
    R : string = '';
    P: Integer;
   begin
     if S = ''
      then Exit( '');
     repeat
       P := Pos( '%', S);
       if P <> 0
        then begin
          R := R + LeftStr( S, P - 1) + Chr( Hex2Dec( Copy( S, P + 1, 2)));
          S := FromStr( S, P + 3)
         end
        else R := R + S
      until P = 0;
     URLDecode := R
   end;

  function URLEncode( S: string): string;
   begin
     URLEncode := URLEncode( S, [])
    end;

  function URLEncode( S: string; Extra: TSysCharSet;
                      ExtraOnly: Boolean = False): string;
   var
    Encodes: TSysCharSet;
    R : string = '';
    P: Integer;
   begin
     if S = ''
      then Exit( '');
     if ExtraOnly
      then Encodes := Extra
      else Encodes := URLEncodes + Extra;
     repeat
       P := PosSet( Encodes, S);
       if P <> 0
        then begin
          R := R + LeftStr( S, P - 1) + '%' + IntToHex( Ord( S[ P]), 2);
          S := FromStr( S, P + 1)
         end
        else R := R + S
      until P = 0;
     URLEncode := R
    end;

  function Wrap( S: string; WrapStyle : TWrapStyle = wsNone;
                 UTF8 : Boolean = True): string;
   const
    Delims = [#0..' ',',','.',';','/','\',':']; // characters used as delimiter
    Unprint = [#$00 .. #$20]; // unprintable characters
    LeadBytes = [#$00 .. #$ff] - [#$80 .. #$bf]; // utf-8 lead bytes incl. ascii
   var
    CharCount, // utf-8 characters counted in line
    AddByteCount, // count of additional bytes in line, being utf-8 followers
    ActualLineWidth, // actual linewidth. differs from linewidth only with urt-8
    LastDelimPos, // position of last delimiter character found
    LastPos, // position of last character in next line
    FirstPos: SizeInt; // position of first character in rest of text
    WrapRes: string;
   procedure AddLine( Line: string; LastLine: Boolean);
    const
     NonSpace = [#$00 .. #$ff] - [' '];
    var
     P, Missing: SizeInt;
    begin
     Missing := LineWidth - Length( Line) - AddByteCount;
     case WrapStyle of
       wsCenter: Line := DupeString( ' ', Missing div 2) + Line;
       wsRight: Line := DupeString( ' ', Missing) + Line;
       wsJustify: if not LastLine
                   then begin
                         P := Pos( ' ', Line);
                         if P <> 0
                          then repeat
                                Insert( ' ', Line, P);
                                Dec( Missing);
                                P := PosSetEx( NonSpace, Line, P);
                                if P = 0
                                 then P := 1;
                                P := PosEx( ' ', Line, P)
                               until Missing = 0
                        end
      end;
     if WrapRes <> ''
      then Line := LineEnding + Line;
     WrapRes := WrapRes + Line
    end;
   begin
    if TabWidth = 0
     then TabWidth := 8;
    if LineWidth = 0
     then LineWidth := 80;
    WrapRes := '';
    repeat
     while (S <> '') and (S[ 1] in UnPrint) do
      S := FromStr( S, 2);
     if S = ''
      then Break;
     if UTF8
      then begin
            ActualLineWidth := 1;
            CharCount := 0;
            LastDelimPos := 0;
            AddByteCount := 0;
            repeat
             if S[ ActualLineWidth] in LeadBytes
              then Inc( CharCount)
              else Inc( AddByteCount);
             Inc( ActualLineWidth);
             if (CharCount >= LineWidth) or (ActualLineWidth >= Length( S))
              then Break;
             if S[ ActualLineWidth] in Delims
              then LastDelimPos := ActualLineWidth
            until False; // endless loop
            if CharCount < LineWidth
             then Break;
            Dec( ActualLineWidth)
           end
      else begin
            ActualLineWidth := LineWidth;
            AddByteCount := 0;
            if Length( S) >= ActualLineWidth
             then Break;
            LastDelimPos := RPosSetEx( Delims, S, ActualLineWidth);
           end;
     if LastDelimPos <> 0
      then begin
            LastPos := LastDelimPos - Ord( S[ LastDelimPos] in Unprint);
            FirstPos := LastDelimPos + 1
           end
      else begin
            LastPos := ActualLineWidth;
            FirstPos := ActualLineWidth + 1
           end;
     AddLine( LeftStr( S, LastPos), False);
     S := FromStr( S, FirstPos);
    until False; // endless loop
    AddLine( S, True);
    Wrap := WrapRes
   end;

  function XEQT( Parent: TComponent; Command: string): string;
   var
    Shell: TProcess;
    TempStrs: TStringList;
    Output: string;
   begin
    Shell := TProcess.Create( Parent);
    if Shell <> nil
     then begin
           Shell.CommandLine := Command;
           Shell.Options := Shell.Options + [poWaitOnExit, poUsePipes];
           TempStrs := TStringList.Create;
           Shell.Execute;
           TempStrs.LoadFromStream( Shell.Output);
           Output := TempStrs.Text;
           while (Output <> '') and (Output[ Length( Output)] in [#10, #13]) do
            Output := LeftStr( Output, Length( Output)- 1);
           XEQT := Output;
           TempStrs.Free
          end;
    Shell.Free
   end;

end.

